/* palmlib.cpp - Palm database file reader
 *
 */

/*
 * Copyright (c) 2002 Tony Aiuto
 *
 * Permission is granted to anyone to use this software for any purpose on
 * any computer system, and to alter it and redistribute it freely, subject
 * to the following restrictions:
 *
 * 1. Neither the authors nor their employers are responsible for
 *    any consequences of the use of this software.
 *
 * 2. The origin of this software must not be misrepresented, either by
 *    explicit claim or by omission.  Credit to the authors must appear
 *    in documentation and sources.
 *
 * 3. Altered versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 *
 * 4. This copyright notice may not be removed or altered.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL TONY AIUTO, HIS CURRENT OR PAST EMPLOYERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/*
 * Author: Tony Aiuto (tony@aiu.to)
 *
 * Edit History:
 * $Log: palmlib.cpp $
 * Revision tools_TonyAiuto/7 2002/09/07 05:48:48 TonyAiuto
 * 	add copyright stuff
 * 
 * Wed Sep  4 23:59:18 EDT 2002
 * tony - created
 */


#ifndef lint
static char _copyright[] = "@(#) Copyright (c) 2002, Tony Aiuto";
static char rcsid[] = "@(#) $Id: palmlib.cpp tools_TonyAiuto/7 2002/09/07 05:48:48 TonyAiuto $";
#endif


#include <stdio.h>
#include <assert.h>
#include <limits.h>

#include "palmlib.h"

char *PalmUtil::phoneTypes[] = {
	"work",
	"home",
	"fax",
	"other",
	"email",
	"main",
	"pager",
	"mobile"
};

int PalmUtil::nPhoneTypes()
{
	return sizeof(PalmUtil::phoneTypes)/sizeof(PalmUtil::phoneTypes[0]);
}

PalmRead::PalmRead(FILE *_input)
{
	input = _input;
	// ateof = 0;
	filename = NULL;
	custheader = NULL;
	categoryCount = 0;
	categories = NULL;
	fieldTypes = NULL;
}

PalmRead::~PalmRead()
{
	if(input != stdin) fclose(input);
	delete [] filename;
	delete [] custheader;
	for(int i = 0; i < categoryCount; i++) {
		delete categories[i];
	}
	delete [] categories;
	delete [] fieldTypes;
}


const char *PalmRead::getCategoryName(int i) const
{
	if(i <= 0) {
		return "unknown";
	} else if(i <= getCategoryCount()) {
		return categories[i-1]->longname;
	} else {
		static char tmp[30];
		sprintf(tmp, "**bad %d **", i);
		return tmp;
	}
}


unsigned char PalmRead::getByte() const
{
	int v = fgetc(input);
	if(v < 0) {
		// ateof = 1;
		return 0;
	}
	return (unsigned char) v;
}

long PalmRead::getShort() const
{
	int a,b;
	a = fgetc(input);
	b = fgetc(input);
	// printf("a = 0x%x, b = %x\n", a, b);
	long x = ((unsigned) (b & 0xff) << 8) | (a & 0xff);
	// printf("x = 0x%x\n", x);
	return x;
}

long PalmRead::getLong() const
{
	short a,b;
	a = getShort();
	b = getShort();
	return ((b & 0xffff) << 16) | (a & 0xffff);
}

char *PalmRead::getCString() const
{
	int len = fgetc(input);
	if(len < 0) return NULL;
	if(len == 0xff) {
		len = getShort();
	}
	if(len < 0) {
		fprintf(stderr, "getString: problem length %d < 0\n", len);
		exit(1);
	}
	char *ret = new char[len + 1];

	char *x = ret;
	while(len-- > 0) {
		*x++ = fgetc(input);
	}
	*x = '\0';
	return ret;
}



PalmRepeat::PalmRepeat()
{
	classname = NULL;
	dateExceptions = NULL;
}

PalmRepeat::~PalmRepeat()
{
	delete [] classname;
	delete [] dateExceptions;
}

void PalmRepeat::load(const PalmRead *reader)
{
	long	ftype = reader->getRawLong();
	assert(ftype == PALM_REPEAT);

	nDateExceptions = reader->getShort();
	if(nDateExceptions > 0) {
		dateExceptions = new long[nDateExceptions];
		for(int ni = 0; ni < nDateExceptions; ni++) {
			dateExceptions[ni] = reader->getLong();
		}
	}
	eventFlag = reader->getShort();
	if(eventFlag == 0) {
		return;
	}

	if(eventFlag == 0xffff) {
		unknown1 = reader->getShort();
		short len = reader->getShort();
		classname = new char[len+1];
		int ci = 0;
		while(len-- > 0) {
			classname[ci++] = reader->getByte();
		}
		classname[ci] = '\0';
	}
	brand = reader->getLong();
	interval = reader->getLong();
	endDate = (time_t) reader->getLong();
	dayOfWeek = reader->getLong();

	// brand data
	if(brand <= 3) {
		day = reader->getLong();
	} 
	if(brand == 2) {
		daymask = reader->getByte();
	}
	if(brand == 3) {
		weeki = reader->getLong();
	}
	if(brand == 4 || brand == 5) {
		day = reader->getLong();
	}
	if(brand == 5) {
		monthi = reader->getLong();
	}
}

long PalmRepeat::dateException(int i)
{
	if(0 <= i && i < nDateExceptions) return dateExceptions[i];
	return 0;
}

void PalmRepeat::print(FILE *out)
{
	if(eventFlag == 0) {
		return;
	}

	fprintf(out, "repeat {");
	if(eventFlag == 0xffff) {
		fprintf(out, "class '%s'", classname);
	} else {
		fprintf(out, "0x%x", eventFlag);
	}
	fprintf(out, " brand 0x%x interval %d", brand, interval);
	if(endDate != 0) {
		fprintf(out, " enddate %s", PalmUtil::time2str(&endDate));
	}
	fprintf(out, " dow %d", dayOfWeek);

	// brand data
	if(brand <= 3) {
		fprintf(out, " day %d", day);
	} 
	if(brand == 2) {
		fprintf(out, " daymask %d", daymask);
	}
	if(brand == 3) {
		fprintf(out, " weeki %d", weeki);
	}
	if(brand == 4 || brand == 5) {
		fprintf(out, " day# %d", day);
	}
	if(brand == 5) {
		fprintf(out, " monthi %d", monthi);
	}
	fprintf(out, "}\n");
	if(nDateExceptions > 0) {
		fprintf(out, "exceptions {");
		for(int i = 0; i < nDateExceptions; i++) {
			fprintf(out, " %ld", dateExceptions[i]);
		}
		fprintf(out, "}\n");
	}
}

void PalmRead::printheader(FILE *out)
{
	printf("#signature: 0x%x = ", signature);
	if(signature == PALM_DBA_SIG) {
		fprintf(out, "datebook\n");
	} else if(signature == PALM_ABA_SIG) {
		fprintf(out, "address\n");
	} else if(signature == PALM_MEMO_SIG) {
		fprintf(out, "memo.dat\n");
	} else if(signature == PALM_TODO_SIG) {
		fprintf(out, "todo.dat\n");
	} else {
		fprintf(stderr, "Not in .dba format (sig = 0x%x)\n", signature);
		return;
	}
	fprintf(out, "#file: %s\n", filename);
	fprintf(out, "#custheader: %s\n", custheader);
	fprintf(out, "#nextFreeCategory: %d\n", nextFreeCategory);
	fprintf(out, "#categoryCount: %d\n", categoryCount);

	for(int i = 0; i < categoryCount; i++) {
		categories[i]->print(out);
	}

	fprintf(out, "#resourceID: %d\n", resourceID);
	fprintf(out, "#fieldsPerRow: %d\n", fieldsPerRow);
	fprintf(out, "#recordIDPos: %d\n", recordIDPos);
	fprintf(out, "#recordStatusPos: %d\n", recordStatusPos);
	fprintf(out, "#placementPos: %d\n", placementPos);

	fprintf(out, "#fieldCount: %d\n", fieldCount);
	for(int i = 0; i < fieldCount; i++) {
		fprintf(out, "#field[%d]: %d\n", i, fieldTypes[i]);
	}
	fprintf(out, "#numEntries: %d\n", numEntries);
}

int PalmRead::loadheader()
{
	int	i;

	signature = getLong();
	if(signature == PALM_DBA_SIG) {
		// OK
	} else if(signature == PALM_ABA_SIG) {
		// OK
	} else if(signature == PALM_MEMO_SIG) {
		// OK
	} else if(signature == PALM_TODO_SIG) {
		// OK
	} else {
		fprintf(stderr, "Not in .dba format (sig = 0x%x)\n", signature);
		return -1;
	}

	filename = getCString();
	custheader = getCString();

	nextFreeCategory = getLong();
	categoryCount = getLong();

	categories = new (PalmCategory*)[categoryCount];
	for(i = 0; i < categoryCount; i++) {
		categories[i] = new PalmCategory(this);
	}

	resourceID = getLong();
	fieldsPerRow = getLong();
	recordIDPos = getLong();
	recordStatusPos = getLong();
	placementPos = getLong();

	fieldCount = getShort();
	fieldTypes = new short[fieldCount];
	for(i = 0; i < fieldCount; i++) {
		fieldTypes[i] = getShort();
	}

	long xxx = getLong();
	numEntries = xxx / fieldsPerRow;
	if(numEntries * fieldsPerRow != xxx) {
		fprintf(stderr, "numEntries (%d) is not a multiple of fieldsPerRow (%d)\n", xxx, fieldsPerRow);
	}

	return 0;
}

long PalmRead::getLongField() const
{
	long ftype = this->getLong();
	assert(ftype == PALM_LONG);
	return this->getLong();
}

char *PalmRead::getCStringField() const
{
	long ftype = this->getLong();
	assert(ftype == PALM_CSTRING);
	ftype = this->getLong();
	assert(ftype == 0);
	return this->getCString();
}

time_t PalmRead::getTimeField() const
{
        long ftype = this->getLong();
        assert(ftype == PALM_TIME);
        return (time_t) this->getLong();
}

long PalmRead::getBoolField() const
{
        long ftype = this->getLong();
        assert(ftype == PALM_BOOL);
        return this->getLong();
}


PalmCategory::PalmCategory(PalmRead *reader)
{
	index = reader->getLong();
	id = reader->getLong();
	dirty = reader->getLong();
	longname = reader->getCString();
	shortname = reader->getCString();
}

PalmCategory::~PalmCategory()
{
	delete [] shortname;
	delete [] longname;
}

void PalmCategory::print(FILE *out)
{
	fprintf(out, "category %d %d %d '%s' '%s'\n", index, id, dirty,
			longname, shortname);
}

PalmEntry *PalmRead::getEntry() const
{
	if(signature == PALM_DBA_SIG) {
		return new DateBookEntry(this);
	} else if(signature == PALM_ABA_SIG) {
		return new AddressBookEntry(this);
	} else if(signature == PALM_MEMO_SIG) {
		return new MemoEntry(this);
	} else if(signature == PALM_TODO_SIG) {
		return new TodoEntry(this);
	} else {
		return NULL;
	}
}


PalmEntry::~PalmEntry()
{
	// nothing to do
}

void PalmEntry::load(const PalmRead *reader)
{
	recordID = reader->getLongField();
	recStatus = reader->getLongField();
	position = reader->getLongField();
}

const char *AddressBookEntry::type() { return "address"; }

AddressBookEntry::AddressBookEntry(const PalmRead *reader)
{
	this->PalmEntry::load(reader);

	lastname = reader->getCStringField();
	firstname = reader->getCStringField();
	title = reader->getCStringField();
	company = reader->getCStringField();

	for(int pi = 0; pi < 5; pi += 1) {
		phoneNumbers[pi] =  new PhoneNumber(reader);
	}

	address = reader->getCStringField();
	city = reader->getCStringField();
	state = reader->getCStringField();
	zip = reader->getCStringField();
	country = reader->getCStringField();
	note = reader->getCStringField();

	isPrivate = reader->getBoolField();
	categoryIndex = reader->getLongField();

	for(int ci=0; ci < 4; ci++) {
		custom[ci] = reader->getCStringField();
	}
	displayIndex = reader->getLongField();

}

AddressBookEntry::~AddressBookEntry()
{
	delete [] lastname;
	delete [] firstname;
	delete [] title;
	delete [] company;

	delete [] address;
	delete [] city;
	delete [] state;
	delete [] zip;
	delete [] country;
	delete [] note;

	for(int i=0; i < 5; i++) {
		delete phoneNumbers[i];
	}
	for(int i=0; i < 4; i++) {
		delete [] (custom[i]);
	}
}



void AddressBookEntry::print(const PalmRead *reader, FILE *out)
{
	fprintf(out, "recordID %d", recordID);
	fprintf(out, " recStatus 0x%02x", recStatus);
	if(position != INT_MAX) {
		fprintf(out, " position %x", position);
	}
	fprintf(out,", category: %s", reader->getCategoryName(categoryIndex));
	PalmUtil::pbool(out, isPrivate, "private");
	fprintf(out, "\n");

	if(*firstname != '\0' || *lastname != '\0') {
		if(*lastname != '\0') fprintf(out, "%s,", lastname);
		if(*firstname != '\0') fprintf(out, "%s", firstname);
		fprintf(out, "\n");
	}
	if(company != NULL && *company != '\0') {
		fprintf(out, "company: %s\n", company);
	}
	if(title != NULL && *title != '\0') {
		fprintf(out, "title: %s\n", title);
	}
	if(address != NULL && *address != '\0') {
		fprintf(out, "%s\n", address);
	}
	if(city != NULL && *city != '\0') {
		fprintf(out, "%s", city);
	}
	if(state != NULL && *state != '\0') {
		fprintf(out, " %s", state);
	}
	if(zip != NULL && *zip != '\0') {
		fprintf(out, "  %s", zip);
	}
	if(country != NULL && *country != '\0') {
		fprintf(out, ", %s", country);
	}
	fprintf(out, "\n");
	for(int pi = 0; pi < 5; pi += 1) {
		fprintf(out, "phone %d: ", pi+1);
		if(displayIndex == pi+1) {
			fprintf(out, "*");
		} else {
			fprintf(out, " ");
		}
		int li = phoneNumbers[pi]->labelIndex;
		char *ptype;
		char tmp[50];
		if(0 <= li && li < PalmUtil::nPhoneTypes()) {
			ptype = PalmUtil::phoneTypes[li];
		} else {
			sprintf(tmp, "(*** %d ***)", li);
			ptype = tmp;
		}
		fprintf(out, "%s %s\n", ptype, phoneNumbers[pi]->text);
	}
	if(*note != '\0') {
		fprintf(out, "note {%s}\n", note);
	}
}

PhoneNumber::PhoneNumber(const PalmRead *reader)
{
	labelIndex = reader->getLongField();
	text = reader->getCStringField();
}

PhoneNumber::~PhoneNumber()
{
	delete [] text;
}


const char *DateBookEntry::type() { return "date"; }

DateBookEntry::DateBookEntry(const PalmRead *reader)
{
	this->PalmEntry::load(reader);

	startTime = reader->getTimeField();
	// Note: weird typing
	endTime = (time_t) reader->getLongField();
	name = reader->getCStringField();
	duration = reader->getLongField();
	note = reader->getCStringField();

	untimed = reader->getBoolField();
	isPrivate = reader->getBoolField();

	categoryIndex = reader->getLongField();
	alarm = reader->getBoolField();
	alarmWarn = reader->getLongField();
	alarmWarnType = reader->getLongField();

	repeat.load(reader);
}

DateBookEntry::~DateBookEntry()
{
	delete [] name;
	delete [] note;
}

void DateBookEntry::print(const PalmRead *reader, FILE *out)
{
	fprintf(out, "id: %08x", recordID);
	fprintf(out, ", status: 0x%x", recStatus);
	if(position != INT_MAX) {
		fprintf(out, ", position: %x", position);
		// fprintf(out, " {%s}", PalmUtil::time2str((time_t *) &position));
	}
	fprintf(out,", category: %s", reader->getCategoryName(categoryIndex));
	fprintf(out, "\n");

	fprintf(out, "{%s}", name);
	fprintf(out, " start %s", PalmUtil::time2str(&startTime));
	fprintf(out, " end %s", PalmUtil::time2str(&endTime));

	PalmUtil::pbool(out, untimed, "untimed");
	PalmUtil::pbool(out, isPrivate, "private");
	PalmUtil::pbool(out, alarm, "alarm");
	if(alarm && alarmWarn != 0) {
		fprintf(out, " warn %d", alarmWarn);
		if(alarmWarnType == 0) {
			fprintf(out, "m");
		} else if(alarmWarnType == 1) {
			fprintf(out, "h");
		} else if(alarmWarnType == 2) {
			fprintf(out, "d");
		} else {
			fprintf(out, " {illegal warntype %d}", alarmWarnType);
		}
	}
	fprintf(out, "\n");

	repeat.print(out);
	if(*note != '\0') {
		fprintf(out, "note {%s}\n", note);
	}
}


const char *MemoEntry::type() { return "memo"; }

MemoEntry::MemoEntry(const PalmRead *reader)
{
	this->PalmEntry::load(reader);
	memo = reader->getCStringField();
	isPrivate = reader->getBoolField();
	categoryIndex = reader->getLongField();
}

MemoEntry::~MemoEntry()
{
	delete [] memo;
}

void MemoEntry::print(const PalmRead *reader, FILE *out)
{
	fprintf(out, "%s", memo);
	fprintf(out,", category: %s", reader->getCategoryName(categoryIndex));
	PalmUtil::pbool(out, isPrivate, "private");
	fprintf(out, "\n");

}


const char *TodoEntry::type() { return "todo"; }

TodoEntry::TodoEntry(const PalmRead *reader)
{
	this->PalmEntry::load(reader);
	description = reader->getCStringField();
	due = reader->getTimeField();
	completed = reader->getBoolField();
	priority = reader->getLongField();
	isPrivate = reader->getBoolField();
	categoryIndex = reader->getLongField();
	note = reader->getCStringField();
}

TodoEntry::~TodoEntry()
{
	delete [] description;
	delete [] note;
}

void TodoEntry::print(const PalmRead *reader, FILE *out)
{
	fprintf(out, "%s", description);
	fprintf(out,", category: %s", reader->getCategoryName(categoryIndex));
	fprintf(out, ", priority %d", priority);
	PalmUtil::pbool(out, isPrivate, "private");
	PalmUtil::pbool(out, completed, "completed");
	if(due != 0) {
		fprintf(out, " due %s", PalmUtil::time2str(&due));
	}
	fprintf(out, "\n");

	if(note != NULL && *note != '\0') {
		fprintf(out, "note: %s\n", note);
	}
}

char *PalmUtil::time2str(time_t *theTime)
{
	static char ret[50];
	struct tm *ts;

	if(*theTime == 0x749d3450) {
		strcpy(ret, "forever");
		return ret;
	}

	ts = gmtime(theTime);
	sprintf(ret, "%d-%02d-%02d %02d:%02d:%02d",
			ts->tm_year + 1900, ts->tm_mon + 1, ts->tm_mday,
			ts->tm_hour, ts->tm_min, ts->tm_sec);
	if(ts->tm_year + 1900 == 2031) {
		sprintf(&ret[strlen(ret)], "=%d", *theTime);
	}
	return ret;
}


void PalmUtil::pbool(FILE *out, long boolval, const char *tag)
{
	if(boolval == 0) return;
	if(boolval == 1) fprintf(out, " %s", tag);
	else fprintf(out, " %s(0x%x)", tag, boolval);
}


void PalmUtil::dumpField(const PalmRead *reader, FILE *out)
{
	long fieldType = reader->getLong();
	if(fieldType == PALM_LONG) {
		long value = reader->getLong();
		fprintf(out, "long %d", value);
	} else if(fieldType == PALM_TIME) {
		time_t value = (time_t) reader->getLong();
		// fprintf(out, "time %d", value);
		fprintf(out, " time %s", time2str(&value));
	} else if(fieldType == PALM_CSTRING) {
		long pad = reader->getLong();
		char * value = reader->getCString();
		fprintf(out, "'%s'", value);
		delete [] value;
	} else if(fieldType == PALM_BOOL) {
		long value = reader->getLong();
		if(value == 0) {
			fprintf(out, "F");
		} else {
			fprintf(out, "T");
		}
	} else if(fieldType == PALM_REPEAT) {
		PalmRepeat	repeat;
		repeat.load(reader);
		repeat.print(out);
	} else {
		fprintf(out, "# UNKNOWN field type %d", fieldType);
	}
}

void PalmUtil::dumpEntry(const PalmRead *reader, FILE *out)
{
	long	ftype;

	/* These first three fields always happen to be this, but
	 * the actual truth is to find the their positions from
	 * the meta-data provided in the schema.   It just always
	 * happens to be the first three fields.
	 */
	ftype = reader->getLong();
	assert(ftype == PALM_LONG);
	long recordID = reader->getLong();

	ftype = reader->getLong();
	assert(ftype == PALM_LONG);
	long recStatus = reader->getLong();

	ftype = reader->getLong();
	assert(ftype == PALM_LONG);
	long position = reader->getLong();

	fprintf(out, "{ recordID %d", recordID);
	fprintf(out, " recStatus 0x%02x", recStatus);
	if(position != INT_MAX) {
		fprintf(out, " position %x", position);
		// fprintf(out, " {%s}", :time2str((time_t *) &position));
	}
	fprintf(out, "\n");

	/* now the varying data */
	for(int fi = 3; fi < reader->fieldCount; fi++) {
		if(fi > 3) fprintf(out, ", ");
		PalmUtil::dumpField(reader, out);
	}
	fprintf(out, "}\n\n");
}

